//
// Probe a host continuously. Log periods of up and down time.
//
char *copyrightNotice = "Copyright (c) 2002 by Michael Newman all rights reserved.\n";
// This application may be used without charge for non-commercial use.
// For other use contact the author at:
//
//	mnewman@dragonnorth.com
//	Michael Newman
//	319 Dean Road
//	Brookline, MA
//	02445
//
// Revision History:
//	 1May2002:	Add brief threshold.
//	26Apr2002:	Defaults enhanced and updated.
//	22Apr2002:	Created

int loggerEdit = 3;


#include <dos.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <winsock.h>

#include "basetype.h"
#include "ping.h"

#define false (1 == 0)
#define true (1 == 1)

LOCALVAR int majorVersion = 1;
LOCALVAR int minorVersion = 0;
LOCALVAR int editNumber;

LOCALVAR int errorCount = 0; // total errors in the program (not ping failures)

// Variables set from the user interface
LOCALVAR int cyclesPerReport = 100;
LOCALVAR int sleepIfBAD = 2;		// Sleep if any nodes were bad
LOCALVAR int sleepIfOK = 10;		// Sleep if all nodes were OK
LOCALVAR int pingCount = 1;
LOCALVAR BOOLEAN reportConnectionLoss = FALSE;
LOCALVAR LPCSTR pLogFileName = NULL;	// Log file name
LOCALVAR FILE *logFile = stdout;	// when above is set from the UI file is opened here
LOCALVAR BOOLEAN verboseLogs = 0;
LOCALVAR int briefThreshold = 1;

// Time the test started
LOCALVAR time_t testStarted;

// These are the hosts to ping
LOCALVAR struct {
    char *pHostName;
    UINT32 hostIP;
    int lastPingOK;
    time_t firstOKtime;
    time_t firstBADtime;
    int numberBrief;			// pings that were bad in a brief sequence
    int numberLong;			// pings that were bad in a long sequence
    int numberBAD;
    int numberOK;
    int firstBADcount;
    int firstOKcount;
    int timeouts;
} hostList[20];
LOCALVAR numHosts = 0;


// Get a time as a string
//
// Inputs:
//	pointer to timestamp
// Returns:
//	pointer to string form of time valid untill next call
static char *timeAsString(time_t *timestamp) {
    static char timeString[100];
    strftime(timeString,
	     sizeof(timeString),
	     "%d%b%Y %H:%M:%S",
	     localtime(timestamp));
    return timeString;
}

// Get time as a string and a timestamp
//
// Inputs:
//	pointer to timestamp
// Returns:
//	pointer to string form of time valid untill next call
static char *nowAsString(time_t *timestamp) {
    time_t tim;
    time(&tim);
    if (timestamp) *timestamp = tim;
    return timeAsString(&tim);
}

// Convert IP address to string
//
// Inputs:
//	ipAddr
// Returns:
//	string ip address valid untill next call or inet_ntoa call
LOCALDEF char *ipToString(UINT32 ipAddr)
{
    struct sockaddr_in sa;
    sa.sin_addr.s_addr = ipAddr;
    sa.sin_family = AF_INET;
    sa.sin_port = 0;
    return inet_ntoa(sa.sin_addr);
}

// Add a host to ping to our list of hosts
// Translate the host name to an IP address
//
// Inputs:
//	hostName
LOCALDEF void addPingHost(char *hostName)
{
    LPHOSTENT lpHost;
    if (numHosts < DIM(hostList)) {
	hostList[numHosts].pHostName = hostName;
	// Lookup host once at start
	lpHost = gethostbyname(hostName);
	if (lpHost == NULL) {
	    printf("Host not found: %s\n", hostName);
	    return;
	};
	hostList[numHosts].hostIP = *((u_long FAR *) (lpHost->h_addr));
	hostList[numHosts].lastPingOK = FALSE;
	hostList[numHosts].numberBrief = 0;	// ignores the first report
	hostList[numHosts].numberLong = 0;
	hostList[numHosts].numberBAD = 0;
	hostList[numHosts].numberOK = 0;
	hostList[numHosts].firstBADcount = 0;
	hostList[numHosts].firstOKcount = 0;
	time(&hostList[numHosts].firstOKtime);
	hostList[numHosts].firstBADtime = hostList[numHosts].firstOKtime;
	hostList[numHosts].timeouts = 0;
	numHosts += 1;
    } else {
	printf("? Too many hosts. Max is %d\n",DIM(hostList));
    };
    return;
}


// Report a change in status of one node.
//
// Inputs:
//	logFile		file to send report to
//	reportNode	TRUE to report a particular target node
//	reportStatus	TRUE to report status of all nodes
//	idx		which target node changed
LOCALDEF void reportTransition(FILE *logFile,BOOLEAN reportNode,BOOLEAN reportStatus,int idx)
{
    time_t tNow;
    time_t deltaTime;	// time in previous state
    int count;		// number of checks in previous state
    int i;
    int totalBAD = 0;
    int totalOK = 0;

    if (!logFile) return;
    if (reportNode) {
	if (hostList[idx].lastPingOK) {
	    fprintf(logFile,"$BACK\t");
	    count = 1 + (hostList[idx].numberBAD - hostList[idx].firstBADcount);
	    deltaTime = hostList[idx].firstOKtime - hostList[idx].firstBADtime;
	} else {
	    if (!reportConnectionLoss) return;
	    fprintf(logFile,"$LOST\t");
	    deltaTime = hostList[idx].firstBADtime - hostList[idx].firstOKtime;
	    count = 1 + (hostList[idx].numberOK - hostList[idx].firstOKcount);
	};
	fprintf(logFile,"%s\t",nowAsString(&tNow));
	fprintf(logFile,"%d\t",count);
	fprintf(logFile,"%d:%02d:%02d\t",(deltaTime/60)/60,(deltaTime/60)%60,deltaTime%60); // hh:mm:ss
	fprintf(logFile,"%s\t",hostList[idx].pHostName);
	fprintf(logFile,"\n");
    };
    if (!reportStatus) return;
    //
    // Report current status of each node as a comment in the log
    //
    fprintf(logFile,"!Status\tOK\tBRIEF\tLONG\tBAD\tRatio\tNode\n");
    for (i = 0;i < numHosts;i += 1) {
	int goodRatio;
	int badRatio;
	if (hostList[i].numberBAD > 0) {
	    goodRatio = hostList[i].numberOK;
	    goodRatio += hostList[i].numberBAD - 1;
	    goodRatio /= hostList[i].numberBAD;
	    badRatio = 1;
	} else {
	    goodRatio = 1;
	    badRatio = 0;
	};
	fprintf(logFile,"! %s\t%d\t%d\t%d\t%d\t%d/%d\t%s\n",
		((hostList[i].lastPingOK) ? "OK" : "BAD"),
		hostList[i].numberOK,
		hostList[i].numberBrief,
		hostList[i].numberLong,
		hostList[i].numberBAD,
		goodRatio,badRatio,
		hostList[i].pHostName);
	totalOK += hostList[i].numberOK;
	totalBAD += hostList[i].numberBAD;
    };
    fprintf(logFile,"!       %d PINGs %d were bad at %s",
	    (totalOK + totalBAD),totalBAD,
	    nowAsString(NULL));
    fprintf(logFile," since %s\n",timeAsString(&testStarted));
    return;
}

// Perform pinging of one host
//
// Inputs:
//	logFile	file to log errors to
//	idx	index to hostList
// Returns:
//	TRUE if all pings were OK
LOCALDEF BOOLEAN performPinging(FILE *logFile,int idx)
{
    BOOLEAN anyBad;
    int result;

    result = ping(logFile,hostList[idx].pHostName,hostList[idx].hostIP,pingCount,verboseLogs,
		  &hostList[idx].timeouts);
    if (result) {
	hostList[idx].numberOK += 1;
	anyBad = FALSE;
    } else {
	hostList[idx].numberBAD += 1;
	anyBad = TRUE;
    };
    if (result != hostList[idx].lastPingOK) {
	hostList[idx].lastPingOK = result;
	if (result) {
	    time(&hostList[idx].firstOKtime);
	    if (hostList[idx].lastPingOK) {
		int count;
		count = 1 + (hostList[idx].numberBAD - hostList[idx].firstBADcount);
		if (hostList[idx].numberBAD > 0) {
		    // Only count if there have been outages to
		    // eliminate counting when first starting up.
		    if (count <= briefThreshold) {
			hostList[idx].numberBrief += count;
		    } else {
			hostList[idx].numberLong += count;
		    };
		};
	    };
	    reportTransition(logFile,TRUE,verboseLogs,idx);
	    hostList[idx].firstOKcount = hostList[idx].numberOK;
	} else {
	    time(&hostList[idx].firstBADtime);
	    reportTransition(logFile,TRUE,verboseLogs,idx);
	    hostList[idx].firstBADcount = hostList[idx].numberBAD;
	};
    };
    return (BOOLEAN)(!anyBad);
}

/* DOCUMENTATION:USER
** The program pings a list of computers and logs when those computers
** are up and down.
**
** Here is the USAGE printed by the program:
** DOCUMENTATION_END
*/
LOCALVAR char *helpMessage =
/* DOCUMENTATION:USER */
			    "LOGGER pings several systems and forms a log of access to those systems.\n"
			    "\n"
			    "Usage is  LOGGER <host> <host>...  \n"
			    "    Arguments start with - everything else is a host to probe.\n"
			    "\n"
			    "    -b# sets threshold of what is considered a brief outage (default 1)\n"
			    "    -c# sets number of pings per host in a row (default 1)\n"
			    "    -d[-] report transitions to not responging (default ON)\n"
			    "    -ofilename sets log file (default is the console)\n"
			    "    -r# sets number of cycles per console summary (default 100)\n"
			    "    -s# seconds to sleep after all OK (default 10)\n"
			    "    -S# seconds to sleep after any BAD (default 1)\n"
			    "    -v# verbose output style 0, 1 (default 1)\n"
			    "\nOnce LOGGER has started type the letter Q and the enter key\n"
			    "to end logging\n\n"
			    "This application may be used without charge for non-commercial use.\n"
			    "For other use contact the author at:\n"
			    "	 mnewman@dragonnorth.com\n"
			    "    Michael Newman,  319 Dean Road,  Brookline, MA 02445\n"
/* DOCUMENTATION_END */
			    ;


// Set in main used at exit
LOCALVAR int nRet;
LOCALVAR int testCycles = 0;

EXTERNDEF void cleanupAtExit(void)
{
    printf("! Ending after %d test cycles at %s\n",testCycles,nowAsString(NULL));
    reportTransition(stdout,FALSE,TRUE,0);
    if (logFile != stdout) {
	fprintf(logFile,"! Ending after %d test cycles at %s\n",testCycles,nowAsString(NULL));
	reportTransition(logFile,FALSE,TRUE,0);
    };
    if (nRet) WSACleanup();		// Free WinSock
    if (logFile != stdout) fclose(logFile); // close log 
    if (errorCount > 0) {
	printf("\n%d errors encountered.\n",errorCount);
    };
    return;
}

/* LOGGER main program
*/
EXTERNDEF int main(int argc,char *argv[]) {
    int i;
    WSADATA wsaData;
    WORD wVersionRequested = MAKEWORD(1,1);

    editNumber = loggerEdit +
		 pingHedit +
		 pingEdit +
		 basetypeHedit;
    printf("\nLOGGER %d(%d)-%d\n",majorVersion,minorVersion,editNumber);
    printf("    %s\n\n",copyrightNotice);
    if (argc < 2) {
	printf(helpMessage);
	exit(EXIT_FAILURE);
    };
    // Init WinSock
    nRet = WSAStartup(wVersionRequested, &wsaData);
    if (nRet) {
	errorCount += 1;
	printf("Error initializing WinSock\n");
	exit(EXIT_FAILURE);
    };
    // Check version
    if (wsaData.wVersion != wVersionRequested) {
	errorCount += 1;
	printf("WinSock version not supported\n");
	WSACleanup();			// Free WinSock
	exit(EXIT_FAILURE);
    };
    errorCount = 0;
    for (i = 1;i < argc;i += 1) {
	if (argv[i][0] == '-') {
	    switch (argv[i][1]) {
		case 'c':
		    pingCount = atoi(&argv[i][2]);
		    break;
		case 'd':
		    reportConnectionLoss = (BOOLEAN)((char)argv[i][2] != '-');
		    break;
		case 'o':
		    pLogFileName = &argv[i][2];
		    if (logFile != stdout) fclose(logFile);
		    logFile = fopen(pLogFileName,"a");
		    if (logFile == NULL) {
			printf("Open file failed: %s\n",pLogFileName);
			logFile = stdout;
			pLogFileName = NULL;
		    };
		    break;
		case 'r':
		    cyclesPerReport = atoi(&argv[i][2]);
		    if (cyclesPerReport <= 0) {
			printf("? Cycles per report was %d must be greater than 0\n",
			       cyclesPerReport);
			cyclesPerReport = 1;
		    };
		    break;
		case 's':
		    sleepIfOK = atoi(&argv[i][2]);
		    break;
		case 'S':
		    sleepIfBAD = atoi(&argv[i][2]);
		    break;
		case 'v':
		    verboseLogs = (BOOLEAN)(atoi(&argv[i][2]) != 0);
		    break;
		default:
		    printf("? Unrecognized selection after -: %s\n",argv[i]);
		    errorCount += 1;
		    break;
	    };
	    continue;
	};
	addPingHost(argv[i]);
    };
    fflush(stdout);
    if (pLogFileName) printf("Log file: %s\n",pLogFileName);
    // Report what we're doing
    printf("Ping checks started at %s\n",nowAsString(&testStarted));
    printf("Pinging %d times per host\n",pingCount);
    printf("%d Cycles per console report\n",cyclesPerReport);
    if (logFile != stdout) {
	fprintf(logFile,"!Ping checks started at %s\n",nowAsString(NULL));
    };
    printf("Hosts to Ping with %d bytes of data:\n",REQ_DATASIZE);
    for (i = 0;i < numHosts;i += 1) {
	fprintf(stdout,"    [%s]\t%s\n",
		ipToString(hostList[i].hostIP),
		hostList[i].pHostName);
    };
    if (logFile != stdout) {
	fprintf(logFile,"!Hosts to Ping with %d bytes of data:\n",REQ_DATASIZE);
	for (i = 0;i < numHosts;i += 1) {
	    fprintf(logFile,"!   [%s]\t\t%s\n",
		    ipToString(hostList[i].hostIP),
		    hostList[i].pHostName);
	};
    };
    printf("\n\nType Q to exit\n\n");
    printf("!what   date      time          since status    node\n");
    printf("!                               changed\n");
    printf("!                               #       time\n");
    atexit(cleanupAtExit);
    while (TRUE) {
	static BOOLEAN reportNow = FALSE;
	int numBadNodes = 0;
	int i;
	for (i = 0;i < numHosts;i += 1) {
	    if (!performPinging(logFile,i)) {
		numBadNodes += 1;
	    };
	};
	testCycles += 1;
	if (reportNow || (cyclesPerReport == 1) || ((testCycles % cyclesPerReport) == 1)) {
	    reportNow = FALSE;
	    printf("! %d cycles completed\n",testCycles);
	    reportTransition(stdout,FALSE,TRUE,0);
	};
#if 1
	if (kbhit()) {
	    int quitNow = FALSE;
	    int c;
	    while (kbhit()) {
		c = fgetc(stdin);
		switch (c) {
		    case 'q':
		    case 'Q':
		    case 'x':
		    case 'X':
			quitNow = TRUE;
			break;
		    case 'r':
		    case 'R':
			reportNow = TRUE;
		    case 'EOF':
		    default:
			quitNow = FALSE;
			break;
		};
		if (quitNow) break;
		if (reportNow) break;
	    };
	    if (quitNow) break;
	};
#endif
	fflush(logFile);
	// Wait each time through the loop
	sleep((numBadNodes) ? sleepIfBAD : sleepIfOK);
    };
    exit(EXIT_SUCCESS);
    return EXIT_SUCCESS;
}


// What happened?
void ReportError(FILE *fp,LPCSTR pWhere)
{
    fprintf(fp,"\n%s error: %d\n",pWhere,WSAGetLastError());
    if ((fp != stderr) && (fp != stdout)) {
	fprintf(stderr,"\n%s error: %d\n",pWhere,WSAGetLastError());
    };
    return;
}
