//
// Ping service
//
// Copyright (c) 2002 by Michael Newman all rights reserved.
// This application may be used without charge for non-commercial use.
// For other use contact the author at:
//
//	mnewman@dragonnorth.com
//	Michael Newman
//	319 Dean Road
//	Brookline, MA
//	02445
//
// Revision History:
//	22April2002:	Created

extern int pingEdit = 1;

#include <time.h>
#include <winsock.h>

#include "basetype.h"
#include "ping.h"

#pragma pack(1)

#define ICMP_ECHOREPLY	0
#define ICMP_ECHOREQ	8

#define false (1 == 0)
#define true (1 == 1)


// IP Header -- RFC 791
typedef struct tagIPHDR
{
    u_char  VIHL;			// Version and IHL
    u_char	TOS;			// Type Of Service
    short	TotLen;			// Total Length
    short	ID;			// Identification
    short	FlagOff;		// Flags and Fragment Offset
    u_char	TTL;			// Time To Live
    u_char	Protocol;		// Protocol
    u_short	Checksum;		// Checksum
    struct	in_addr iaSrc;		// Internet Address - Source
    struct	in_addr iaDst;		// Internet Address - Destination
}IPHDR, *PIPHDR;


// ICMP Header - RFC 792
typedef struct tagICMPHDR
{
    u_char	Type;			// Type
    u_char	Code;			// Code
    u_short	Checksum;		// Checksum
    u_short	ID;			// Identification
    u_short	Seq;			// Sequence
    char	Data;			// Data
}ICMPHDR, *PICMPHDR;



// ICMP Echo Request
typedef struct tagECHOREQUEST
{
    ICMPHDR icmpHdr;
    DWORD	dwTime;
    char	cData[REQ_DATASIZE];
}ECHOREQUEST, *PECHOREQUEST;


// ICMP Echo Reply
typedef struct tagECHOREPLY
{
    IPHDR	ipHdr;
    ECHOREQUEST	echoRequest;
    char    cFiller[256];
}ECHOREPLY, *PECHOREPLY;


#pragma pack()



// Checksum routine for Internet Protocol family headers (C Version)
// The algorithm is simple, using a 32 bit accumulator (sum),
// add sequential 16 bit words to it, and at the end, fold
// back all the carry bits from the top 16 bits into the lower 16 bits.
//
static u_short in_cksum(u_short *addr, int len)
{
    register int nleft = len;
    register u_short *w = addr;
    register u_short answer;
    register int sum = 0;

    while( nleft > 1 )  {
	sum += *w++;
	nleft -= 2;
    };

    // mop up an odd byte, if necessary
    if( nleft == 1 ) {
	u_short	u = 0;

	*(u_char *)(&u) = *(u_char *)w ;
	sum += u;
    };

    // add back carry outs from top 16 bits to low 16 bits
    sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
    sum += (sum >> 16);			/* add carry */
    answer = (u_short)(~sum);		/* truncate to 16 bits */
    return (answer);
}

// SendEchoRequest()
// Fill in echo request header
// and send to destination
static int SendEchoRequest(SOCKET s,LPSOCKADDR_IN lpstToAddr) 
{
    static ECHOREQUEST echoReq;
    static short nId = 1;
    static short nSeq = 1;
    int nRet;

    // Fill in echo request
    echoReq.icmpHdr.Type		= ICMP_ECHOREQ;
    echoReq.icmpHdr.Code		= 0;
    echoReq.icmpHdr.Checksum	= 0;
    echoReq.icmpHdr.ID			= nId++;
    echoReq.icmpHdr.Seq			= nSeq++;

    // Fill in some data to send
    for (nRet = 0; nRet < REQ_DATASIZE; nRet++)
	echoReq.cData[nRet] = (char)(' ' + nRet);

    // Save tick count when sent
    echoReq.dwTime				= GetTickCount();

    // Put data in packet and compute checksum
    echoReq.icmpHdr.Checksum = in_cksum((u_short *)&echoReq, sizeof(ECHOREQUEST));

    // Send the echo request  								  
    nRet = sendto(s,			/* socket */
		  (LPSTR)&echoReq,	/* buffer */
		  sizeof(ECHOREQUEST),
		  0,			/* flags */
		  (LPSOCKADDR)lpstToAddr, /* destination */
		  sizeof(SOCKADDR_IN));	/* address length */

    if (nRet == SOCKET_ERROR) 
	ReportError(stderr,"sendto()");
    return (nRet);
}


// RecvEchoReply()
// Receive incoming data
// and parse out fields
static DWORD RecvEchoReply(SOCKET s, LPSOCKADDR_IN lpsaFrom, u_char *pTTL) 
{
    ECHOREPLY echoReply;
    int nRet;
    int nAddrLen = sizeof(struct sockaddr_in);

    // Receive the echo reply	
    nRet = recvfrom(s,			// socket
		    (LPSTR)&echoReply,	// buffer
		    sizeof(ECHOREPLY),	// size of buffer
		    0,			// flags
		    (LPSOCKADDR)lpsaFrom, // From address
		    &nAddrLen);		// pointer to address len

    // Check return value
    if (nRet == SOCKET_ERROR) 
	ReportError(stderr,"recvfrom()");

    // return time sent and IP TTL
    *pTTL = echoReply.ipHdr.TTL;
    return(echoReply.echoRequest.dwTime);   		
}


// WaitForEchoReply()
// Use select() to determine when
// data is waiting to be read
static int WaitForEchoReply(SOCKET s)
{
    struct timeval Timeout;
    fd_set readfds;

    readfds.fd_count = 1;
    readfds.fd_array[0] = s;
    Timeout.tv_sec = 5;		// wait up to 5 seconds
    Timeout.tv_usec = 0;

    return(select(1, &readfds, NULL, NULL, &Timeout));
}

// Probe a host with a ping request n times. Report failures
//
// Calls SendEchoRequest() and
// RecvEchoReply() and prints results
//
// Inputs:
//	fp		file to report failures to
//	pstrHost	name of host for error reporting
//	hostIP		host to probe
//	count		number of times to probe
//	verbose		0 for no logs of timeouts
//			1 to log timeouts
//			2 also reports time of each ping
// Outputs:
//	timouts		pointer to cell counted for each timeout
// Returns:
//	true	host was present and answered
//	false	no host or no answer
int ping(FILE *fp,LPCSTR pstrHost,UINT32 hostIP,int count,int verbose,int *timeouts)
{
    SOCKET	  rawSocket;
    struct    sockaddr_in saDest;
    struct    sockaddr_in saSrc;
    DWORD	  dwTimeSent;
    DWORD	  dwElapsed;
    u_char    cTTL;
    int       nLoop;
    int       nRet;
    int	numberErrors;

    // Create a Raw socket
    rawSocket = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
    if (rawSocket == SOCKET_ERROR) {
	ReportError(stderr,"socket()");
	return false;
    };

    // Setup destination socket address
    saDest.sin_addr.s_addr = hostIP;
    saDest.sin_family = AF_INET;
    saDest.sin_port = 0;

    // Ping multiple times
    numberErrors = 0;
    for (nLoop = 0; nLoop < count; nLoop++)
    {
	// Send ICMP echo request
	SendEchoRequest(rawSocket, &saDest);

	// Use select() to wait for data to be received
	nRet = WaitForEchoReply(rawSocket);
	if (nRet == SOCKET_ERROR)
	{
	    ReportError(stderr,"select()");
	    numberErrors += 1;
	    break;
	};
	if (nRet == 0) {
	    // Timeout occured. Log it and do not try any more.
	    if (timeouts) *timeouts = *timeouts + 1;
	    if (verbose > 0) {
		fprintf(fp,"!TimeOut %s\n",pstrHost);
	    };
	    numberErrors += 1;
	    break;
	};

	// Receive reply
	dwTimeSent = RecvEchoReply(rawSocket, &saSrc, &cTTL);

	// Calculate elapsed time
	dwElapsed = GetTickCount() - dwTimeSent;
	if (verbose > 1) {
	    printf("Reply from: %s: bytes=%d time=%ldms TTL=%d\n", 
		   inet_ntoa(saSrc.sin_addr), 
		   REQ_DATASIZE,
		   dwElapsed,
		   cTTL);
	};
    };
    nRet = closesocket(rawSocket);
    if (nRet == SOCKET_ERROR) {
	ReportError(stderr,"closesocket()");
    };
    return (numberErrors == 0);
}
